from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from era.utils.functools import unidec, omit


@unidec
def role_required(method, req, *args, **kw):
    if req.user.role in kw.get('allow', []):
        return method(req, *args, **omit(kw, 'allow'))
    raise PermissionDenied()



from .templatetags.library import register, Component, TemplateComponent, Tag, ScriptedTag
from .templatetags.markup import Icon, MarkedList, IconicList, Break, ProgressBar, \
    Row, Container, Column, Link, Button, Label, Alert, Well, Caption, Panel, Accordion, \
    Navbar, Table
from era.templatetags.media import Script, Stylesheet, Image, Messages
from .templatetags.menu import Menu, Dropdown
from .templatetags.forms import Form
from .templatetags.lists import Paginator, ChangeList, SearchLine



from functools import reduce
from itertools import chain
from django.conf import settings
from django.apps import apps
from django.utils.text import capfirst

from ..utils.functools import factual, first, pick, omit
from ..utils.translation import get_string, get_model_names
from ..utils.urls import exists_import
from .library import register, Component, ComplexComponent, Tag
from .markup import Link, Caption


class MenuItem(Tag):
    el = 'li'

    def get_defaults(self):
        return {
            'divider': False,
            'include': True,
            'disabled': False,
            'link': {'url': None}}

    def get_nodelist(self):
        if self.props.divider or not self.props.include:
            return ''
        return self.inject(Link, self.props.link, self.props.caption)

    def resolve_props(self):
        return {'caption': self.inject(Caption, self.props.caption)}

    def tweak(self):
        if not self.props.include:
            self.dom.empty()
        elif self.props.divider:
            self.dom.add_class('nav-divider')
        else:
            self.dom.add_class(self.get_class_set('disabled', 'active'))
        super().tweak()


class Dropdown(Tag):
    el = 'li'

    def get_defaults(self):
        return {'active': False}

    def resolve_props(self):
        result = {}
        if self.props.active:
            result['class'] = 'active'
        if self.props.nodelist:
            if not 'caret' in self.props:
                result['caret'] = self.inject(
                    Tag, {'el': 'span', 'class': 'caret'})
        else:
            result['caret'] = ''
        return result

    def get_nodelist(self):
        return ''.join([
            self.inject(
                Link,
                self.props.link,
                ''.join([
                    self.inject(Caption, self.props.caption),
                    self.props.caret]),
                attrs={
                    'class': 'dropdown-toggle',
                    'data-toggle': 'dropdown'}),
            '' if not self.props.nodelist else self.inject(
                Tag,
                {'el': 'ul', 'class': 'dropdown-menu'},
                self.props.nodelist)])


class Menu(Tag):
    el = 'ul'
    inline = True

    def get_items(self):
        return self.props.get('items', [])

    def resolve_item(self, item):
        link = item.get('link', {})
        if 'divider' in item:
            return {}
        elif 'model' in item:
            model = apps.get_model(*item['model'].split('.'))
            if not 'title' in item['caption']:
                item['caption']['title'] = get_model_names(model)[-1]
            (prefix, url) = map(get_string, get_model_names(model))
            if not 'url' in link:
                link['url'] = url
            item['active'] = self.check_active(
                url, *map(
                    lambda suffix: '-'.join([prefix, suffix]),
                    ['add', 'edit']))
        elif not 'active' in item:
            if link.get('url') and link.get('reverse', True):
                item['active'] = self.check_active(link.get('url'))
            else:
                item['active'] = False
        if 'include' in item and item['include'] is None:
            item['include'] = item.get('active')
        if item.get('disabled'):
            link.update({'url': '#', 'reverse': False})
        return dict(item, link=link)

    def resolve_items(self):
        return map(self.resolve_item, self.get_items())

    def filter_items(self, fn=lambda i: True, items=None):
        if items is None:
            items = self.resolve_items()
        return list(filter(lambda i: i.get('include', True) and fn(i), items))

    def render_item(self, item):
        if 'dropdown' in item:
            return self.render_dropdown(item) or ''
        return self.inject(MenuItem, item)

    def render_items(self, items=None):
        if items is None:
            items = self.resolve_items()
        return ''.join(map(self.render_item, items))

    def render_dropdown(self, item):
        items = list(item['dropdown'].resolve_items())
        include_items = self.filter_items(items=items)
        active_items = self.filter_items(lambda i: i.get('active'), items)

        if len(include_items):
            toggle = item['caption'].get('toggle', True)
            props = dict({
                'link': {},
                'nodelist': '',
                'active': bool(active_items)},
                **pick(item, 'caption', 'attrs', 'link', 'caret'))

            display = first(props['active'] and active_items or include_items)
            if item['caption'].get('collapse', True) and len(include_items) == 1:
                if not props['active']:
                    props['link'].update(pick(display, 'url'))
                if toggle:
                    props.update(pick(display, 'caption'))
                return self.inject(MenuItem, props)
            if toggle and props['active']:
                props.update(pick(display, 'caption'))
                props['nodelist'] = self.render_items(self.filter_items(
                    lambda i: not i.get('active'),
                    items))
            else:
                props['nodelist'] = self.render_items(items)
            return self.inject(Dropdown, props)

    def get_defaults(self):
        return {'pills': False, 'tabs': False, 'stacked': False}

    def resolve_attrs(self):
        return {'class': self.get_class_set(
            'pills', 'tabs', 'stacked', prefix='nav', include='nav')}

    def get_nodelist(self):
        return self.render_items()


@register.era
class MainMenu(Menu):
    def get_items(self):
        return list(chain(*map(
            lambda cls: self.insert(cls).get_items(),
            factual(map(
                lambda module: getattr(
                    module,
                    ''.join([capfirst(module.__name__.split('.')[0]), 'Menu']),
                    None),
                factual(map(
                    lambda app: exists_import('.'.join([app, 'components'])),
                    getattr(settings, 'MAIN_MENU', []))))))))



import json
from django import forms
from django.conf import settings
from django.templatetags.static import static
from django.utils import translation
from .utils.functools import first, avg


class EmptyWidget(forms.widgets.Widget):
    def render(self, *args, **kw):
        return ''


class Slider(forms.widgets.TextInput):
    class Media:
        css = {'all': [static('seiyria-bootstrap-slider/dist/css/bootstrap-slider.min.css')]}
        js = [
            static('seiyria-bootstrap-slider/dist/bootstrap-slider.min.js'),
            static('widgets/slider.js')]

    def __init__(self, attrs=None):
        attrs = attrs or {}
        if not 'range' in attrs:
            attrs['range'] = range(0, 100, 1)
        attrs.update(dict(zip(('min', 'max', 'step'), map(
            lambda a: getattr(attrs['range'], a),
            ('start', 'stop', 'step')))))
        attrs['value'] = attrs.pop('value', int(
            avg(attrs['range'].start, attrs.pop('range').stop)))
        super().__init__(dict(map(
            lambda t: ('data-slider-' + t[0], t[1]),
            attrs.items())))


class FrozenSelect(forms.widgets.Select):
    def render(self, name, value, attrs=None, choices=()):
        return ''.join([
            str(list(filter(lambda c: c[0] == value, self.choices))[0][1]),
            forms.widgets.HiddenInput().render(name, value)])


class DateTimePicker(forms.widgets.DateTimeBaseInput):
    class Media:
        css = {'all': [static(
            'eonasdan-bootstrap-datetimepicker/build/css/bootstrap-datetimepicker.min.css')]}
        js = [static(
            'eonasdan-bootstrap-datetimepicker/build/js/bootstrap-datetimepicker.min.js')]

    def __init__(self, key='date', attrs=None, options=None):
        self.options = options or {}
        self.set_format(key)
        super().__init__(attrs)

    def set_format(self, key):
        self.format_key = '_'.join([key, 'input_formats']).upper()
        self.options['format'] = ''.join(map(
            lambda c: {
                'j': 'DDD',
                'd': 'DD',
                'B': 'MMMM',
                'b': 'MMM',
                'm': 'MM',
                'Y': 'YYYY',
                'y': 'YY',
                'H': 'HH',
                'I': 'hh',
                'M': 'mm',
                'S': 'ss',
                'p': 'a',
                'z': 'ZZ'} \
                .get(c, c),
            getattr(settings, '_'.join([key, 'format']).upper())))

    def get_icons(self):
        return {
            'time': 'fa fa-clock-o',
            'date': 'fa fa-calendar',
            'up': 'fa fa-arrow-up',
            'down': 'fa fa-arrow-down',
            'next': 'fa fa-arrow-right',
            'previous': 'fa fa-arrow-left'}

    def render(self, name, value, attrs):
        return ''.join([
            super().render(name, value, dict(attrs, readonly='readonly')),
            '<script>$("#{0}").datetimepicker({1});</script>'.format(
                attrs['id'], json.dumps(dict({
                    'icons': self.get_icons(),
                    'locale': translation.get_language(),
                    'ignoreReadonly': True,
                    'stepping': 5
                    }, **(self.options or {}))))])



import string

from django.conf import settings
from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.auth.decorators import login_required
from django.contrib.auth.hashers import get_hasher, make_password

from era import _, random_str
from era.views import RedirectView, FormView, CollectionView, ObjectView
from era.utils.functools import factual, pick

from .components import JoinNotification, ResetNotification, InviteNotification
from .decorators import role_required
from .forms import LoginForm, ProfileForm, JoinForm, ResetForm, ConfirmForm, UserForm
from .models import Confirm


class PasswordMixin:
    def gen_password(self, **kw):
        return random_str(5, ''.join([
            string.ascii_uppercase,
            string.digits]), **kw)

    def set_password(self, form):
        if form.cleaned_data['password']:
            form.instance.set_password(form.cleaned_data['password'])


class UserMixin(PasswordMixin):
    def get_user(self, form):
        model = get_user_model()
        return model.objects.filter(**pick(form.cleaned_data, model.USERNAME_FIELD)).first()


class ConfirmationMixin:
    def gen_code(self, salt=None):
        hasher = get_hasher()
        generate = True
        while generate:
            code = random_str()
            encoded = hasher.encode(code, salt or hasher.salt())
            generate = bool(Confirm.objects.filter(code=encoded).count())
        return code, encoded

    def create_confirmation(self, user, key, password):
        code, encoded = self.gen_code(password)
        Confirm.objects.create(
            user=user,
            key=key,
            code=encoded,
            sign=make_password(password))
        return code


class LoginMixin:
    def get_access(self, user):
        return user.access

    def process_login(self, user):
        if self.get_access(user):
            auth.login(self.request, user)
            return self.success_finish()
        else:
            self.send_message('error', _('sorry, access denied'))
            return self.navigate('login')

    def get_success_message(self, **kw):
        return _('welcome {username}').format(username=str(self.request.user))


class HistoryNavigationMixin:
    back_button_url = 'index'

    def get_back_action(self):
        return {
            'icon': 'chevron-left',
            'title': _('back'),
            'level': 'warning',
            'link': self.back_button_url}


class AuthFormMixin:
    keywords = ['auth']
    form_props = {'class': 'condensed'}


class AuthRequestView(
    UserMixin, ConfirmationMixin, HistoryNavigationMixin, AuthFormMixin, FormView):
    long_term = True
    back_button_url = 'login'


class LoginView(LoginMixin, AuthFormMixin, FormView):
    form_class = LoginForm

    @property
    def has_comm(self):
        return get_user_model().get_communicator(user=None) is not None

    def get_actions(self):
        result = [{
                'icon': 'sign-in',
                'title': _('login'),
                'level': 'success'}]
        if self.has_comm:
            result.extend([{
                'icon': 'user-plus',
                'title': _('join'),
                'level': 'default',
                'link': 'join'
            }, {
                'icon': 'unlock',
                'title': _('unlock'),
                'level': 'danger',
                'link': 'reset'
            }])
        return result

    def process_valid(self, form, **kw):
        user = auth.authenticate(**form.cleaned_data)
        if user:
            if Confirm.objects.filter(user=user, key='registration').count():
                self.send_message('error', _('sorry, unconfirmed data'))
            else:
                return self.process_login(user)
        else:
            self.send_message('error', _('sorry, invalid credentials'))
        return self.reload()


class LogoutView(RedirectView):
    decorators = [login_required]
    permanent = True

    def get_redirect_url(self):
        auth.logout(self.request)
        self.pattern_name = 'index'
        return super().get_redirect_url()


class ProfileView(UserMixin, LoginMixin, FormView):
    decorators = [login_required]
    form_class = ProfileForm
    form_props = {'class': 'condensed'}

    def get_instance(self):
        return self.request.user

    def process_valid(self, form, **kw):
        if form.cleaned_data['password']:
            self.set_password(form)
            return self.process_login(auth.authenticate(**dict(
                self.request.user.username_dict,
                password=form.cleaned_data['password'])))
        return super().process_valid(**dict(kw, form=form))

    def get_success_message(self, **kw):
        return _('your profile was updated successfully')


class JoinView(AuthRequestView):
    form_class = JoinForm
    success_redirect = 'confirm'
    success_message = _('confirmation code has been sent')

    def get_actions(self):
        return [{
            'icon': 'check',
            'title': _('submit'),
            'level': 'success'
        }, self.get_back_action()]

    def process_valid(self, form, **kw):
        user = self.get_user(form)
        if user:
            confirm = Confirm.objects.filter(user=user, key='registration').first()
            if confirm:
                user.delete()
                confirm.delete()
            else:
                self.send_message('error', _('sorry, user with such data exists'))
                return self.reload()
        return super().process_valid(form=form, **kw)

    def save_form(self, form):
        form.instance.role = settings.USER_ROLES[-1].string
        self.set_password(form)
        super().save_form(form)
        form.instance.comm.send(
            self.request,
            _('registration'),
            JoinNotification,
            code=self.create_confirmation(
                form.instance,
                'registration',
                form.cleaned_data['password']))


class ResetView(AuthRequestView):
    form_class = ResetForm
    success_redirect = 'unlock'
    success_message = _('confirmation code has been sent')
    actions = [{'icon': 'send', 'title': _('request'), 'level': 'success'}]

    def process_valid(self, form, **kw):
        user = self.get_user(form)
        if user:
            Confirm.objects.filter(user=user, key='password').delete()
            user.comm.send(
                self.request,
                _('access restoration'),
                ResetNotification,
                code=self.create_confirmation(
                    user,
                    'password',
                    form.cleaned_data['password']))
            return self.success_finish()
        else:
            self.send_message('error', _('sorry, invalid credentials'))
            return self.reload()


class ConfirmView(LoginMixin, AuthFormMixin, FormView):
    form_class = ConfirmForm
    repeat_url = 'join'
    update_password = False

    def get_initial(self):
        return pick(self.request.GET, 'code')

    def get_actions(self):
        return [
            {'icon': 'check', 'title': _('submit'), 'level': 'success'},
            {'icon': 'refresh', 'title': _('repeat'), 'level': 'default', 'link': self.repeat_url}]

    def process_valid(self, form, **kw):
        confirm = Confirm.objects \
            .filter(code=get_hasher().encode(
                form.cleaned_data['code'],
                form.cleaned_data['sign'])) \
            .first()
        if confirm:
            if self.update_password:
                confirm.user.password = confirm.sign
                confirm.user.save()
            user = confirm.user
            confirm.delete()
            return self.process_login(auth.authenticate(**dict(
                user.username_dict,
                password=form.cleaned_data['sign'])))
        else:
            self.send_message('error', _('sorry, invalid credentials'))
        return self.reload()


class UnlockView(ConfirmView):
    repeat_url = 'reset'
    update_password = True


class UserManagerMixin:
    decorators = [login_required, role_required(allow=['developer'])]
    model = get_user_model()


class BaseUsersView(UserManagerMixin, CollectionView):
    list_display = (get_user_model().USERNAME_FIELD, 'role', 'name', 'access')
    list_filter = ('role', 'access')
    list_counters = ('role', )
    list_search = ('name', )
    default_state = {'filters': {'access': True}}


class BaseUserView(PasswordMixin, UserManagerMixin, ObjectView):
    form_class = UserForm
    form_props = {'class': 'condensed'}

    def send_invite(self, form, password):
        return form.instance.comm.send(
            self.request,
            _('invitation'),
            InviteNotification,
            password=password)

    def save_form(self, form):
        if not form.instance.pk:
            password = form.cleaned_data['password'] or self.gen_password()
            form.instance.set_password(password)
            form.instance.comm and self.send_invite(form, password)
        else:
            self.set_password(form)
        super().save_form(form)


class UsersView(BaseUsersView): pass
class UserView(BaseUserView): pass



from django import forms
from django.contrib.auth import get_user_model
from era import _, random_str
from .models import Confirm


user_field = get_user_model().USERNAME_FIELD
new_password_input = forms.PasswordInput(attrs={'placeholder': _('enter new password')})


class NonUniqueMixin:
    def validate_unique(self):
        pass


class LoginForm(NonUniqueMixin, forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = ('email', 'password')
        widgets = {'password': forms.PasswordInput()}


class ProfileForm(forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = (user_field, 'name', 'password')
        widgets = {
            'email': forms.TextInput(attrs={'readonly': 'readonly'}),
            'password': new_password_input}


class JoinForm(NonUniqueMixin, forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = (user_field, 'name', 'password')
        widgets = {'password': forms.PasswordInput()}


class ResetForm(NonUniqueMixin, forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = (user_field, 'password')
        widgets = {'password': new_password_input}


class ConfirmForm(forms.ModelForm):
    class Meta:
        model = Confirm
        fields = ('code', 'sign')
        widgets = {'sign': forms.PasswordInput(
            attrs={'placeholder': _('enter password again')})}


class UserForm(forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = (user_field, 'role', 'name', 'password', 'access')

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.fields['password'].required = False
        if self.instance.pk:
            self.fields['password'].widget = new_password_input
        else:
            self.fields['password'].widget = forms.PasswordInput({
                'placeholder': _('leave blank for random')})



from itertools import chain, groupby
from django.templatetags.static import static
from django.utils.text import capfirst

from ..utils.functools import just, factual
from .library import register, Component, Tag
from .markup import Alert


class Include(Tag):
    el = 'link'
    nobody = True
    named = False

    def get_defaults(self):
        return {'static': True, 'host': False}

    def get_url(self):
        return ''.join(factual([
            self.props.host and self.get_host(),
            (self.props.static and static or just)(self.props.url)]))

    def resolve_attrs(self):
        return {'rel': self.rel, 'href': self.get_url()}


class Source(Include):
    def resolve_attrs(self):
        return {'src': self.get_url()}


@register.era
class Favicon(Include):
    rel = 'shortcut icon'


@register.era
class Stylesheet(Include):
    rel = 'stylesheet'


@register.era
class Script(Source):
    el = 'script'
    nobody = False
    inline = True

    def get_nodelist(self):
        return ''
        

@register.era
class Image(Source):
    el = 'img'


@register.era
class Messages(Component):
    def get_defaults(self):
        return {
            'capfirst': True,
            'dismiss': True,
            'levels': {
                10: 'primary',
                20: 'info',
                25: 'success',
                30: 'warning',
                40: 'danger'}}

    def DOM(self):
        return self.inject(
            Tag, {'class': 'messages'}, ''.join(map(
                lambda message: self.inject(
                    Alert, {
                        'level': self.props.levels.get(message.level),
                        'dismiss': self.props.dismiss},
                    (self.props.capfirst and capfirst or just)(message.message)),
                self.context['messages'])))



from ..utils.functools import unidec, pluck, separate, pick, omit, truthful, avg
from .base import SimpleTestCase, IsOkTestCase


class UnidecTestCase(IsOkTestCase):
    def test_default_behaviour(self):
        self.assertOk(
            unidec(lambda fn, a: fn(a) + 'k') \
            (lambda a: a[1:])('oo'))

    def test_with_params(self):
        self.assertOk(
            unidec(lambda fn, a, **f: f['w'] + fn(a))(w='o') \
            (lambda a: a[1:])('kk'))


class AvgTestCase(SimpleTestCase):
    def test(self):
        self.assertEqual(avg(4, 6, 8), 6)


class SeparateTestCase(SimpleTestCase):
    def test(self):
        [even, odd] = separate(lambda x: bool(x % 2), [1, 2, 3, 4, 5])
        self.assertEqual(even, [2, 4])
        self.assertEqual(odd, [1, 3, 5])


class DictCopyTestCase(SimpleTestCase):
    def test_pick(self):
        self.assertEqual(
            pick({0: 1, 2: 4}, 0), {0: 1})

    def test_omit(self):
        self.assertEqual(
            omit({0: 1, 2: 4}, 0), {2: 4})

    def test_truthful(self):
        self.assertEqual(
            truthful({1: True, 2: False, 3: 'yes', 4: []}), {1: True, 3: 'yes'})


class PluckTestCase(SimpleTestCase):
    def test_dict(self):
        self.assertEqual(
            pluck([{0: 0}, {0: 1, 1: 1}, {0: 2, 1: 1, 2: 2}], 0),
            [0, 1, 2])

    def test_obj(self):
        class O:
            def __init__(self, x):
                self.x = x

        self.assertEqual(
            pluck([O(1), O(2), O(3)], 'x'),
            [1, 2, 3])



import operator
from functools import reduce
from itertools import chain
from urllib.parse import urlencode

from django.conf import settings
from django.db.models import Count, Q
from django.db.models.fields import BooleanField, FieldDoesNotExist
from django.db.models.fields.related import ManyToOneRel
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.generic.list import BaseListView

from ..components import ChangeList
from ..utils.functools import just, call, first, pluck, pick, \
    map_keys, map_values, reduce_dict, filter_dict
from ..utils.translation import _, get_string, get_model_names, verbose_choices
from .base import BaseView


class ListView(BaseView, BaseListView):
    list_display = []

    @property
    def columns(self, *args, **kw):
        if not hasattr(self, '_columns'):
            self._columns = self.get_list_view('display')
        return self._columns

    def get_column_key(self, column):
        if isinstance(column, str):
            return column
        return column.string.replace(' ', '_')

    def lookup_field(self, obj, field, next_obj=getattr, end_obj=getattr):
        if not '__' in field:
            return end_obj(obj, field)
        return self.lookup_field(
            next_obj(obj, field.split('__')[0]),
            '__'.join(field.split('__')[1:]),
            next_obj=next_obj,
            end_obj=end_obj)

    def get_model_field(self, field):
        try:
            return self.lookup_field(
                self.model,
                get_string(field),
                next_obj=lambda model, field: model._meta.get_field(field).rel.to,
                end_obj=lambda model, field: model._meta.get_field(field))
        except FieldDoesNotExist:
            return {'name': field}

    def get_list_view(self, op, value=None):
        return value or getattr(self, 'list_' + op)

    def lookup_field_display(self, obj, field, model=True):
        method = 'get_{0}_display'.format(field)
        if isinstance(obj, self.model) and hasattr(self, method):
            return getattr(self, method)(obj)
        elif model:
            return call(getattr(obj, method, lambda: getattr(obj, field)))
        raise AttributeError('Missing display data for "{0}"'.format(field))

    def display_column(self, field):
        if isinstance(field, dict):
            return field['name']
        elif isinstance(field, ManyToOneRel):
            return field.field.model._meta.verbose_name_plural
        else:
            return field.verbose_name

    def display_field(self, name, field, obj):
        if isinstance(field, dict):
            return self.lookup_field_display(
                obj,
                get_string(field['name']).replace(' ', '_'),
                model=False)
        elif '__' in get_string(name):
            return self.lookup_field(obj, name, end_obj=self.lookup_field_display)
        else:
            value = self.lookup_field_display(obj, field.name)
            if isinstance(field, ManyToOneRel) and not isinstance(value, str):
                return value.count()
            return value

    def display_objects(self, objects):
        return list(map(
            lambda obj: dict({'pk': obj.pk}, **dict(map(
                lambda c: (
                    self.get_column_key(c),
                    self.display_field(c, self.get_model_field(c), obj)),
                self.columns))),
            objects))


class CollectionView(ListView):
    components = {'content': ChangeList}
    list_filter = []
    list_counters = []
    list_sort = []
    list_search = []
    search_method = 'icontains'
    default_state = {}
    paginate_by = 15
    actions = [{
        'icon': 'plus-square',
        'title': _('Add'),
        'level': 'success',
        'link': {'rel': 'add'}}]

    @property
    def stateful(self):
        if not hasattr(self, '_stateful'):
            self._stateful = self.check_stateful()
        return self._stateful

    def check_stateful(self):
        return True

    def map_attr(self, attr):
        return attr.replace('__', '.')

    def map_state(self, key, method='filter'):
        return '-'.join([method, key])

    def get_default_state(self):
        return self.default_state

    def get_state(self, method):
        return map_keys(
            lambda k: k.replace('.', '__'),
            map_values(
                lambda v: v if not v in ('True', 'False') else v[0] == 'T',
                filter_dict(
                    lambda k, v: k in list(getattr(self, '_'.join([method, 'keys']))),
                    map_keys(
                        lambda k: k.split('-')[1],
                        filter_dict(
                            lambda k, v: k.startswith(method),
                            self.request.GET)))))

    def get_queryset(self, ignore_state=False, ignore_attrs=None):
        qs = self.model.objects.all() if self.queryset is None else self.queryset
        if not ignore_state and self.stateful:
            if self.states['filter']:
                qs = qs.filter(**filter_dict(
                    lambda k, v: not k in (ignore_attrs or []),
                    self.states['filter']))
            if 'search' in self.request.GET:
                for word in self.request.GET['search'].split():
                    qs = qs.filter(reduce(
                        operator.or_, map(
                        lambda lookup: Q(**{lookup: word}),
                        map(
                            lambda f: '__'.join([f, self.search_method]),
                            self.get_list_view('search')))))
            if not ignore_attrs and self.states['sort']:
                qs = qs.order_by(*reduce_dict(
                    lambda k, v: ('-' if v is False else '') + k,
                    self.states['sort']))
        return qs

    def resolve_generic_filter(self, attr, choices, state):
        result = {'key': self.map_attr(attr)}
        if choices:
            result['choices'] = []
            values = pluck(
                self.get_queryset(ignore_state=(not state), ignore_attrs=[attr]),
                result['key'])
            for choice in choices:
                count = values.count(choice[0])
                if count:
                    result['choices'].append(list(chain(choice, [count])))
        return result

    def resolve_filter(self, attr, state):
        field = self.get_model_field(attr)
        if isinstance(field, BooleanField):
            result = self.resolve_generic_filter(
                attr,
                state is not None and [(True, _('Yes').lower()), (False, _('No').lower())],
                state)
        elif field.is_relation:
            result = self.resolve_generic_filter(
                attr + '__pk',
                state is not None and map(
                    lambda obj: (obj.pk, str(obj)),
                    field.rel.to.objects.all()),
                state)
        elif hasattr(field, 'choices'):
            result = self.resolve_generic_filter(
                attr,
                state is not None and field.choices,
                state)
        else:
            method = 'resolve_{0}_filter'.format(get_string(attr))
            result = getattr(self, method)(attr, state=state)
        return result

    def resolve_filters(self, **kw):
        return list(map(
            lambda name: dict({'name': name}, **self.resolve_filter(name, **kw)),
            self.get_list_view('filter')))

    def get_filters(self):
        return list(map(
            lambda f: dict(f, **{
                'title': self.display_column(self.get_model_field(f['name'])),
                'counters': f['name'] in self.get_list_view('counters')}),
            self.resolve_filters(state=True)))

    def get_actions(self):
        return self.actions

    def change_list_view(self, op, value=None):
        if self.stateful and op == 'display':
            value = list(filter(
                lambda f: not f in self.active_filters,
                (value or self.list_display)))
        return value

    def get_list_view(self, op, value=None):
        if not hasattr(self, '_list_view'):
            self._list_view = {}
        if not op in self._list_view:
            self._list_view[op] = super().get_list_view(
                op, self.change_list_view(op, value))
        return self._list_view[op]

    def define_state(self):
        self.sort_keys = list(map(
            self.map_attr,
            self.get_list_view('sort')))
        self.filter_keys = list(map(
            lambda f: f['key'],
            self.resolve_filters(state=None)))
        self.states = {k: self.get_state(k) for k in ('filter', 'sort')}
        self.active_filters = list(map(
            first,
            filter(
                lambda z: self.map_state(z[-1]) in self.request.GET,
                zip(self.get_list_view('filter'), self.filter_keys))))

    @property
    def redirection(self):
        if self.stateful:
            default_state = self.get_default_state()
            if not self.model.objects.count():
                return self.navigate('-'.join([
                    get_string(get_model_names(self.model)[0]),
                    'add']))
            elif not self.request.GET and default_state:
                kw = {}
                if 'filters' in default_state:
                    for f in self.resolve_filters(state=False):
                        if len(f['choices']) == 1:
                            kw[self.map_state(f['key'])] = f['choices'][0][0]
                        elif f['key'] in default_state['filters']:
                            kw[self.map_state(f['key'])] = default_state['filters'][f['key']]
                if 'sort' in default_state:
                    kw.update(map_keys(
                        lambda k: self.map_state(k, 'sort'),
                        self.default_state['sort']))
                if kw:
                    return redirect('?'.join([self.request.path, urlencode(kw)]))
            self.define_state()

    def get(self, request, *args, **kw):
        return self.redirection or super().get(request, *args, **kw)

    def get_guide(self):
        return list(map(
            lambda name: [
                self.get_column_key(name),
                self.display_column(self.get_model_field(name)),
                name in self.get_list_view('sort') and self.map_attr(name)],
            self.columns))

    def get_context_data(self, **kw):
        data = super().get_context_data(**kw)
        if self.stateful:
            data.update({
                'guide': self.get_guide(),
                'filters': self.get_filters(),
                'actions': self.get_actions(),
                'search': bool(len(self.get_list_view('search')))})
        return dict(data, objects=self.display_objects(data['object_list']))



from django.test.runner import DiscoverRunner


class NoDbDiscoverRunner(DiscoverRunner):
    def setup_databases(self, **kwargs):
        pass

    def teardown_databases(self, *args, **kwargs):
        pass
